from core.core30_context.context_dependency_graph import context_producer, context_dependencies
from core.core00_core_model.mixin.base_mixin import BaseMixinsNoJoin
from core.core00_core_model.concept.timed import CreatedModifiedAt
from core.core11_config.config import config_dependencies, Config
from core.core30_context.context import Context

from sqlalchemy import Column, Integer, String, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column
from contextlib import contextmanager
from threading import Thread
import random
import socket
import select
import time


class LocalPortForward(*BaseMixinsNoJoin, CreatedModifiedAt):
    __tablename__ = 'local_port_forward'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    reason: Mapped[str] = mapped_column(String, nullable=True)

    local_ip: Mapped[str] = Column(String, default='127.0.0.1')
    local_port: Mapped[int] = Column(Integer)
    remote_ip: Mapped[str] = mapped_column(String, nullable=False)
    remote_port: Mapped[int] = Column(Integer)

    __table_args__ = (UniqueConstraint("reason", "local_ip", "local_port", "remote_ip", "remote_port",
                                       name="unique_lpf"),)

    def __repr__(self):
        return f"[.] LPF |{self.local_ip}:{self.local_port} -> {self.remote_ip}:{self.remote_port}| ({self.reason})"


class LocalPortForwardManager:
    per_host_port = {}

    def __init__(self, remote_ip: str, remote_port: int, reason: str | None = None,
                 local_ip: str | None = None, local_port: int | None = None):
        self.remote_ip = remote_ip
        self.remote_port = remote_port

        self.local_ip = local_ip or '127.0.0.1'
        self.local_port = local_port or random.randint(20000, 60000)
        assert local_port or reason, 'At least one of reason or local port must be provided to local port forward'

        local_ip_path = str(local_ip).replace('.', '-')
        self.local_ip_path = local_ip_path
        base_path = f".systemctxt.controlled.{local_ip_path}.local_port_forwards.{self.local_port}"

        if base_path in LocalPortForwardManager.per_host_port:
            for m in LocalPortForwardManager.per_host_port[base_path].__dict__:
                setattr(self, m, LocalPortForwardManager.per_host_port[base_path].__dict__[m])
            return

        @context_producer((f".systemctxt.controlled.{local_ip_path}.local_port_forward_manager",
                           LocalPortForwardManager), )
        def manager_for_ip(ctxt: Context):
            ctxt.setdefault('systemctxt', {}).setdefault('controlled', {}) \
                .setdefault(local_ip_path, {})['local_port_forward_manager'] = self

        LocalPortForwardManager.per_host_port[base_path] = self

    # TODO: add parameters for timeout, simultaneous clients, ...
    @contextmanager
    def with_ssh_forward(self, remote_ip: str, remote_port: int):
        lpf = LocalPortForward.query_for(reason=self.reason, remote_ip=remote_ip, remote_port=remote_port,
                                         local_ip=self.local_ip, local_port=self.local_port).one_or_none()
        lpf = lpf or LocalPortForward.create(reason=self.reason, remote_ip=remote_ip, remote_port=remote_port,
                                             local_ip=self.local_ip, local_port=self.local_port)

        server_socket = socket.socket()
        server_socket.settimeout(10)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind((self.local_ip, self.local_port))
        server_socket.listen(10)

        server_finished = False

        def tunnel_to_remote():
            try:
                client_socket, client_address = server_socket.accept()
                client_socket.settimeout(10)
                remote_ip, _ = client_socket.getpeername()
                remote_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                remote_socket.settimeout(10)
                remote_socket.connect((self.remote_ip, self.remote_port))

                start = time.time()
                finished = False
                while not finished and not server_finished:
                    socket_list = [client_socket, remote_socket]
                    read_sockets, write_sockets, error_sockets = select.select(socket_list, [], [])
                    for sock in read_sockets:
                        if sock == client_socket:
                            data = sock.recv(1024)
                            if not data:
                                finished = True
                                break
                            else:
                                remote_socket.send(data)
                        elif sock == remote_socket:
                            data = sock.recv(1024)
                            if not data:
                                finished = True
                                break
                            else:
                                client_socket.send(data)
                    if time.time() - start > 30:
                        print("[.] Time-outed (more than 30s)")
                        break
                    time.sleep(0.1)
            except Exception as e:
                print(f"[-] Unable to receive data: {e}")
            finally:
                try:
                    client_socket.close()
                except:
                    pass
                try:
                    remote_socket.close()
                except:
                    pass

        t = Thread(target=tunnel_to_remote)
        t.start()
        try:
            yield self
        finally:
            print(f"[.] Stop listening Local port forward {lpf}")
            server_socket.close()
            server_finished = True
            t.join()


@context_producer(('.systemctxt.controlled_local.local_port_forward_manager', LocalPortForwardManager))
@context_dependencies(('.targetctxt.step', str), )
@config_dependencies(('.systemctxt.controlled_local.local_forward_port', int), )
def get_local_port_forward_manager(config: Config, ctxt: Context):
    local_port = config['systemctxt']['controlled_local']['local_forward_port']
    local_port = (local_port > 0 and local_port) or None
    lpfm = LocalPortForwardManager(local_port=local_port, reason=ctxt['targetctxt']['step'])
    ctxt['systemctxt']['controlled_local']['local_port_forward_manager'] = lpfm
    return lpfm
