from core.core30_context.context_dependency_graph import context_producer, context_dependencies
from core.core00_core_model.mixin.base_mixin import BaseMixinsNoJoin
from core.core00_core_model.concept.timed import CreatedModifiedAt
from core.core11_config.config import config_dependencies, Config
from core.core02_model.typed.file import FileContent
from core.core30_context.context import Context

from action.utils.control_server import RemoteSSHManager

from ..interact.temporary_server import temporary_server

from sqlalchemy import Column, Integer, String, Boolean, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column
from contextlib import contextmanager
import os.path
import random
import socket
import string


class ReversePortForward(*BaseMixinsNoJoin, CreatedModifiedAt):
    __tablename__ = 'reverse_port_forward'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    reason: Mapped[str] = mapped_column(String, nullable=True)

    remote_ip: Mapped[str] = mapped_column(String, nullable=False)
    remote_port: Mapped[int] = Column(Integer)
    local_ip: Mapped[str] = Column(String, default='127.0.0.1')
    local_port: Mapped[int] = Column(Integer)

    checked_ok: Mapped[bool] = Column(Boolean, default=False)
    service_installed: Mapped[bool] = Column(Boolean, default=False)
    service_running: Mapped[bool] = Column(Boolean, default=False)

    __table_args__ = (UniqueConstraint("reason", "remote_ip", "remote_port", name="unique_rpf"),)

    def __repr__(self):
        return f"[.] RPF |{self.remote_ip}:{self.remote_port} -> {self.local_ip}:{self.local_port}| (checked: " \
               f"{self.checked_ok}, service installed: {self.service_installed}, " \
               f"service running: {self.service_running})"


socat_service = """[Unit]
Description=Socat PORTR
ConditionFileIsExecutable=/usr/bin/socat
After=network.target

[Service]
StartLimitInterval=5
StartLimitBurst=10
ExecStart=/usr/bin/socat TCP4-LISTEN:PORTR,bind=0.0.0.0,fork,reuseaddr TCP4:IPL:PORTL
User=root
Restart=always
RestartSec=60

[Install]
WantedBy=multi-user.target
"""


class ReversePortForwardManager:

    per_host_port = {}

    def __init__(self, ssh_url, local_ip: str | None = None, reason: str | None = None,
                 remote_port: int | None = None, local_port: int | None = None):
        ssh_manager = RemoteSSHManager(ssh_url)
        remote_ip = str(ssh_manager.remote_ip)

        local_ip = local_ip or '127.0.0.1'
        local_port = local_port or random.randint(20000, 60000)
        assert remote_port or reason, 'At least one of reason or remote port must be provided to reverse port forward'
        self.rpf = (remote_port and ReversePortForward.query_for(reason=reason, remote_ip=remote_ip,
                                                                 remote_port=remote_port) or
                    ReversePortForward.query_for(reason=reason, remote_ip=remote_ip)).one_or_none()
        remote_port = remote_port or (self.rpf and self.rpf.remote_port) or random.randint(20000, 60000)
        self.rpf = self.rpf or ReversePortForward.create(reason=reason, remote_ip=remote_ip, remote_port=remote_port,
                                                         local_ip=local_ip, local_port=local_port)
        local_port = self.rpf.local_port
        self.service_name = f"revssh_{remote_port}.service"

        remote_ip_path = str(remote_ip).replace('.', '-')
        self.remote_ip_path = remote_ip_path
        base_path = f".systemctxt.controlled.{remote_ip_path}.reverse_port_forwards.{remote_port}"


        if base_path in ReversePortForwardManager.per_host_port:
            for m in ReversePortForwardManager.per_host_port[base_path].__dict__:
                setattr(self, m, ReversePortForwardManager.per_host_port[base_path].__dict__[m])
            return

        @context_producer((f".systemctxt.controlled.{remote_ip_path}.reverse_port_forward_manager",
                           ReversePortForwardManager),)
        def manager_for_ip(ctxt: Context):
            ctxt.setdefault('systemctxt', {}).setdefault('controlled', {}) \
                .setdefault(self.remote_ip_path, {})['reverse_port_forward_manager'] = self


        @context_dependencies((f".systemctxt.controlled.{remote_ip_path}.ssh_ok", bool),
                              (f".systemctxt.controlled.{remote_ip_path}.ssh_manager", RemoteSSHManager),)
        def get_ssh_manager(ctxt: Context):
            return ctxt['systemctxt']['controlled'][remote_ip_path]['ssh_manager']
        ssh_manager = get_ssh_manager()  # ensure SSH access
        uname = ssh_manager.ssh_service.identity_proof.identity.username


        @context_producer((f"{base_path}.uploaded", bool))
        def upload_reverse_port_forward_service(ctxt: Context):
            ctxt.setdefault('systemctxt', {}).setdefault('controlled', {}).setdefault(remote_ip_path, {}) \
                .setdefault('reverse_port_forwards', {}).setdefault(str(remote_port), {})['uploaded'] = True
            if self.rpf.service_installed:
                return True
            print("[.] Checking RPF service uploaded")

            fcontent = socat_service.replace('PORTR', str(remote_port)) \
                .replace('PORTL', str(local_port)) \
                .replace('IPR', remote_ip) \
                .replace('IPL', local_ip) \
                .replace('USER', uname)
            out_absname = ssh_manager.upload(FileContent(content=fcontent.encode()), self.service_name)
            fname = os.path.basename(out_absname)
            move_cmd_out = ssh_manager.execute_command(f"sudo cp {out_absname} /etc/systemd/system/{fname}")
            reload_systemctl_out = ssh_manager.execute_command('sudo systemctl daemon-reload')
            self.rpf.service_installed = move_cmd_out['status'] == 0 and reload_systemctl_out['status'] == 0

            ctxt['systemctxt']['controlled'][remote_ip_path]['reverse_port_forwards'][str(remote_port)]['uploaded'] = \
                self.rpf.service_installed
            if move_cmd_out['status'] != 0:
                raise Exception(f"[-] CP Command failed: {move_cmd_out['stderr']}({move_cmd_out['status']})")
            if reload_systemctl_out['status'] != 0:
                raise Exception(f"[-] daemon-reload failed: {reload_systemctl_out['stderr']}"
                                f"({reload_systemctl_out['status']})")
            return self.rpf.service_installed


        @context_producer((f"{base_path}.enabled", bool))
        @context_dependencies((f"{base_path}.uploaded", bool))
        def enable_reverse_port_forward_service(ctxt: Context):
            ctxt['systemctxt']['controlled'][remote_ip_path]['reverse_port_forwards'][str(remote_port)]['enabled'] = \
                True
            if self.rpf.service_running:
                return True
            print("[.] Checking RPF enabled")
            enable_cmd_out = ssh_manager.execute_command(f"sudo systemctl enable {self.service_name}")
            start_cmd_out = ssh_manager.execute_command(f"sudo systemctl start {self.service_name}")
            ctxt['systemctxt']['controlled'][remote_ip_path]['reverse_port_forwards'][str(remote_port)]['enabled'] = \
                True
            self.rpf.service_running = enable_cmd_out['status'] == 0 and start_cmd_out['status'] == 0
            return self.rpf.service_running


        @context_producer((f"{base_path}.checked", bool))
        @context_dependencies((f"{base_path}.enabled", bool))
        def check_setup_reverse_port_forward_service(ctxt: Context):
            ctxt['systemctxt']['controlled'][remote_ip_path]['reverse_port_forwards'][str(remote_port)]['checked'] = \
                True
            if self.rpf.checked_ok:
                return True
            print("[.] Checking setup RPF")
            checked = check_reverse_port_forward_service()
            ctxt['systemctxt']['controlled'][remote_ip_path]['reverse_port_forwards'][str(remote_port)]['checked'] = \
                checked
            self.rpf.checked_ok = checked
            return checked


        def check_reverse_port_forward_service():
            class DataChecker:
                def __init__(self):
                    self.random_data = ''.join([random.choice(string.ascii_lowercase) for _ in range(0x30)])
                    self.checked = False

                def check_random_data(self, received, *args, **kwargs):
                    self.checked = self.random_data == received.decode()
                    if self.checked:
                        print("[+] Successfully confirmed remote port forwarding working")
                    else:
                        print("[-] Remote port forwarding failed")

            with ssh_manager.with_ssh_forward(local_port, local_ip, local_port) as _:
                dc = DataChecker()
                with temporary_server(local_ip, local_port, on_recv=dc.check_random_data) as srv:
                    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    try:
                        s.connect((remote_ip, remote_port))
                        s.send(dc.random_data.encode())
                    except:
                        return False
                    finally:
                        s.close()
                        srv.join()
            return dc.checked


        @contextmanager
        @context_dependencies((f"{base_path}.checked", bool),)
        def with_ssh_forward(ctxt: Context):
            with ssh_manager.with_ssh_forward(local_port, local_ip, local_port) as fwd:
                yield fwd


        self.upload_reverse_port_forward_service = upload_reverse_port_forward_service
        self.enable_reverse_port_forward_service = enable_reverse_port_forward_service
        self.check_setup_reverse_port_forward_service = check_setup_reverse_port_forward_service
        self.with_ssh_forward = with_ssh_forward

        ReversePortForwardManager.per_host_port[base_path] = self


    def invalidate_redo(self):
        self.rpf.checked_ok = False
        checked = self.check_setup_reverse_port_forward_service()
        if not checked:
            self.rpf.service_running = False
            checked = self.enable_reverse_port_forward_service()
            checked = checked and self.check_setup_reverse_port_forward_service()
            if not checked:
                self.rpf.service_running = False
                self.rpf.service_installed = False
                checked = self.upload_reverse_port_forward_service()
                checked = checked and self.enable_reverse_port_forward_service()
                checked = checked and self.check_setup_reverse_port_forward_service()
        return checked


@context_producer(('.systemctxt.controlled.reverse_port_forward_manager', ReversePortForwardManager))
@context_dependencies(('.targetctxt.step', str),)
@config_dependencies(('.systemctxt.controlled.reverse_forward_port', int), ('.systemctxt.controlled.server_url', str))
def get_reverse_port_forward_manager(config: Config, ctxt: Context):
    remote_port = config['systemctxt']['controlled']['reverse_forward_port']
    remote_port = (remote_port > 0 and remote_port) or None
    #ip_address = ctxt['systemctxt']['controlled']['ssh_service'].authenticated_object.endpoint.ip_address
    rpfm = ReversePortForwardManager(config['systemctxt']['controlled']['server_url'],
                                     remote_port=remote_port, reason=ctxt['targetctxt']['step'])
    ctxt['systemctxt']['controlled']['reverse_port_forward_manager'] = rpfm
    return rpfm


if __name__ == '__main__':
    from core.core05_persistent_model.policy.session import get_session
    from core.core30_context.context import current_ctxt

    with get_session():
        rpfm = get_reverse_port_forward_manager()
        ctxt = current_ctxt()['systemctxt']['controlled']
        rpf_manager = ctxt['reverse_port_forward_manager']
        checked = rpf_manager.check_setup_reverse_port_forward_service()
        if not checked:
            print(f"[.] Failed to check RPF, invalidate and redo")
            rpf_manager.invalidate_redo()

        ctxt = current_ctxt()['systemctxt']['controlled']
        print(ctxt)
