from construct import Enum, Struct, Bytes, Byte, Array, GreedyString, GreedyBytes, this, Int64ul, Int16ul, Int16sl, Int8ul, Float64l


# Enums

Cmd = Enum(Int16ul,
    AUTHENTICATION_PDU = 0x0001,
    AUTHENTICATION_RESPONSE_PDU = 0x0002,
    HEARTBEAT_PDU = 0x0003,
    SERVER_PLAYER_INFO_PDU = 0x0004,
    CLIENT_PLAYER_INFO_PDU = 0x0005,
    CLIENT_ATTACK_PDU = 0x0006,
    TEXT_MESSAGE_PDU = 0x0007,
    MAP_STATE_PDU = 0x0008,
    OBJECT_SEIZING_PDU = 0x0009,
    CLIENT_DISCONNECT_PDU = 0x000A,
    SERVER_DISCONNECT_PDU = 0x000B,
    SERVER_ATTACK_PDU = 0x000C,
    CHAT_PDU = 0x1000,
    ERROR_PDU = 0x8000,
)

Team = Enum(Int8ul,
    Blue = 0x0,
    Red = 0x1,
)

FacingDirection = Enum(Int8ul,
    Left = 0x0,
    Right = 0x1,
)

ObjectType = Enum(Int8ul,
    Heart = 0x01,
    Greenshard = 0x02,
)


# PDU definition

PDU = Struct(
    "cmdId" / Cmd,
    "bodySize" / Int16ul,
    "token" / Bytes(16),
    "body" / Array(this.bodySize, Byte),
)


# PDU sub-types

AuthenticationPDU = Struct(
    "license_key" / Bytes(32),
    "username" / Bytes(16),
)

AuthenticationResponsePDU = Struct(
    "player_id" / Int16ul,
    "token" / Bytes(16),
    "team" / Team,
)

HeartbeatPDU = Struct()

ServerPlayerInfoPDU = Struct(
    "player_id" / Int16ul,
    "team" / Team,
    "name" / Bytes(16),
    "facing" / FacingDirection,
    "hp" / Int16sl,
    "map_id" / Int8ul,
    "x" / Float64l,
    "y" / Float64l,
    "is_afk" / Int8ul,
    "shielded" / Int8ul,
    "greenshard_hp" / Int64ul,
    "last_attacker" / Bytes(16),
)

ClientPlayerInfoPDU = Struct(
    "name" / Bytes(16),
    "facing" / FacingDirection,
    "map_id" / Int8ul,
    "x" / Float64l,
    "y" / Float64l,
    "is_afk" / Int8ul,
    "shielded" / Int8ul,
    "greenshard_hp" / Int64ul,
    "last_attacker" / Bytes(16),
)

ClientAttackPDU = Struct(
    "victim_player_id" / Int16ul,
    "attack_damage" / Int64ul,
)

TextMessagePDU = Struct(
    "text" / GreedyString("utf8"),
)

MapStatePDU = Struct(
    "n_objects" / Int16ul,
    "objects" / Array(
        this.n_objects,
        Struct(
            "object_id" / Int16ul,
            "map_id" / Int8ul,
            "pos_x" / Int16ul,
            "pos_y" / Int16ul,
            "object_type" / ObjectType,
        )
    ),
)

ObjectSeizingPDU = Struct(
    "object_id" / Int16ul,
)

ClientDisconnectPDU = Struct()

ServerDisconnectPDU = Struct(
    "player_id" / Int16ul,
)

ServerAttackPDU = Struct(
    "perpetrator_id" / Int16ul,
    "victim_id" / Int16ul,
    "damage" / Int64ul,
    "is_dead" / Int8ul,
    "respawn_map_id" / Int8ul,
    "respawn_x" / Int16ul,
    "respawn_y" / Int16ul,
)

ChatPDU = Struct(
    "text" / GreedyBytes,
)

ErrorPDU = Struct(
    "error" / GreedyString("utf8"),
)


# Command mapping

commands = {
    Cmd.AUTHENTICATION_PDU: AuthenticationPDU,
    Cmd.AUTHENTICATION_RESPONSE_PDU: AuthenticationResponsePDU,
    Cmd.HEARTBEAT_PDU: HeartbeatPDU,
    Cmd.SERVER_PLAYER_INFO_PDU: ServerPlayerInfoPDU,
    Cmd.CLIENT_PLAYER_INFO_PDU: ClientPlayerInfoPDU,
    Cmd.CLIENT_ATTACK_PDU: ClientAttackPDU,
    Cmd.TEXT_MESSAGE_PDU: TextMessagePDU,
    Cmd.MAP_STATE_PDU: MapStatePDU,
    Cmd.OBJECT_SEIZING_PDU: ObjectSeizingPDU,
    Cmd.CLIENT_DISCONNECT_PDU: ClientDisconnectPDU,
    Cmd.SERVER_DISCONNECT_PDU: ServerDisconnectPDU,
    Cmd.SERVER_ATTACK_PDU: ServerAttackPDU,
    Cmd.CHAT_PDU: ChatPDU,
    Cmd.ERROR_PDU: ErrorPDU,
}
