import threading

from core.core05_persistent_model.policy.session import get_session
from core.core11_config.config import config_dependencies, Config

from .protocol import Cmd, PDU, commands

from typing import TypeAlias, Tuple, Container, Callable
import socket
import select
import time


LICENSE_KEY = b'PR2YU5CZGCYMS272GLZ1WA43W7P44I7S'


def ensure_service(ip, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    cmd = Cmd.HEARTBEAT_PDU
    serialized_heartbeat = commands[cmd].build({})
    serialized_pdu = PDU.build({
        "cmdId": cmd,
        "bodySize": len(serialized_heartbeat),
        "token": b'\x00'*0x10,
        "body": serialized_heartbeat,
    })
    sock.sendto(serialized_pdu, (ip, port))

    start = time.time()
    while time.time() - start < 6:
        sock.sendto(serialized_pdu, (ip, port))
        ready_to_read, _, _ = select.select([sock], [], [], 0.3)
        if sock in ready_to_read:
            data, addr = sock.recvfrom(1024)
            try:
                pdu: Container = PDU.parse(data)
                print(f"[.] Received heartbeat pdu, server OK")
                sock.close()
                return True
            except:
                pass
    sock.close()
    return False


@config_dependencies(('.systemctxt.local_or_remote', bool))
def pwn_service(config: Config):
    if config['systemctxt']['local_or_remote']:
        from wussticcommon.ensure_service.local_instance import get_working_context as get_working_local

        def check_ctxt(local_sstic_ctxt):
            return ensure_service(local_sstic_ctxt.ip, local_sstic_ctxt.port)
        ctxt = get_working_local(check_ctxt)
        if not check_ctxt(ctxt):
            raise Exception(f"Unable to validate local context {ctxt}")
        return ctxt.ip, ctxt.port
    else:
        from wussticcommon.ensure_service.instance_lifecycle import get_working_context as get_working_remote

        def check_ctxt(remote_sstic_ctxt):
            return ensure_service(remote_sstic_ctxt.ip, remote_sstic_ctxt.port1)
        ctxt = get_working_remote(check_ctxt)
        if not check_ctxt(ctxt):
            raise Exception(f"Unable to validate remote context {ctxt}")
        return ctxt.ip, ctxt.port1


with get_session():
    HOST, PORT = pwn_service()

print(f"[.] Will do for {HOST} {PORT}")

Addr: TypeAlias = Tuple[str, int]
PduHandler: TypeAlias = Callable[[Container], None]

log_state = {}  # prevent displaying the same message received by different player
lock = threading.Lock()
def print_opti(log_line):
    with lock:
        curtime = time.time()
        if log_line in log_state:
            if curtime - log_state[log_line] > 5:  # ok for displaying again
                print(log_line)
        else:
            print(log_line)
        log_state[log_line] = curtime



def send_pdu(sock, cmd: Cmd, body: dict, token: bytes = b"\x00" * 16) -> None:
    serialized_body = commands[cmd].build(body)
    # print(f"Sending {cmd!r}, body ({len(serialized_body)} bytes): {serialized_body.hex()}")

    assert 0 <= len(serialized_body) <= 0xFFFF

    serialized_pdu = PDU.build({
        "cmdId": cmd,
        "bodySize": len(serialized_body),
        "token": token,
        "body": serialized_body,
    })

    sock.sendto(serialized_pdu, (HOST, PORT))


def handle_heartbeat_pdu(body: Container, ctxt):
    print_opti("HEARTBEAT PDU, ignoring")


def handle_client_player_info_pdu(body: Container, ctxt):
    if body.greenshard_hp > 0x14:
        ctxt['leaked'] = body.greenshard_hp
        if body.greenshard_hp not in ctxt.get('all_ever_leaked', set()):
            print_opti(f"[+] Got a new gracefully sent leak: {hex(body.greenshard_hp)}")
        ctxt.setdefault('all_ever_leaked', set()).add(body.greenshard_hp)

    if 'target_player' not in ctxt and body.player_id != ctxt.get('player_id'):
        print_opti(f"[+] We found our target :) {body.player_id}")
        ctxt['target_player'] = body.player_id

    if not ctxt.get('all_players') or body.player_id not in ctxt['all_players']:
        print_opti(f"[+] We found another player entering the ring ! {body.player_id}")
    ctxt.setdefault('all_players', {})[body.player_id] = body



def handle_object_seizing_pdu(body: Container, ctxt):
    print_opti(body)


def handle_client_attack_pdu(body: Container, ctxt):
    print_opti(f"[.] Client {body.victim_id} attack {hex(body.damage)}")

def handle_client_disconnect_pdu(body: Container, ctxt):
    print_opti(f"[.] Client disconnected {hex(body.player_id)}")
    print_opti(body.player_id)


def handle_chat_pdu(body: Container, ctxt):
    print_opti(f"[.] Received chat {body.text.hex()}")


def handle_error_pdu(body: Container, ctxt):
    print_opti(f"[.] Received ERROR {body.error}")


def handle_map_state_pdu(body: Container, ctxt):
    ctxt['n_objects'] = body.n_objects
    ctxt['objects'] = body.objects


def handle_text(body: Container, ctxt):
    print_opti(f"[.] New text: {body.text}")


pdu_handlers: dict[int, PduHandler] = {
    Cmd.HEARTBEAT_PDU: handle_heartbeat_pdu,
    Cmd.SERVER_PLAYER_INFO_PDU: handle_client_player_info_pdu,
    Cmd.SERVER_ATTACK_PDU: handle_client_attack_pdu,
    Cmd.OBJECT_SEIZING_PDU: handle_object_seizing_pdu,
    Cmd.SERVER_DISCONNECT_PDU: handle_client_disconnect_pdu,
    Cmd.CHAT_PDU: handle_chat_pdu,
    Cmd.ERROR_PDU: handle_error_pdu,
    Cmd.MAP_STATE_PDU: handle_map_state_pdu,
    Cmd.TEXT_MESSAGE_PDU: handle_text,
}


if __name__ == '__main__':
    from time import sleep

    from recv_thread import Loop

    l = Loop()
    sleep(5)
    l.end()
