import random
import threading

import protocol



MAX_OBJECTS = 16

RED_BASE_ID = 0x00
MAIN_ARENA_ID = 0x01
BLUE_BASE_ID = 0x02


# `probability` is the probability of apparition during a server tick (`UPDATE_PERIOD`)
# `area` describes the coordinates (in tiles) at which the object may spawn inside main arena
# `max_amount` is the max amount of this object type that can exist on the map at the same time
# `disappearance_probability` is the probability of disappearance during a server tick

OBJECTS = {
    protocol.ObjectType.Heart: {
        "probability": 0.008,
        "area": [
            (11, 12), (12, 12), (13, 12), (14, 12), (15, 12), (16, 12), (17, 12), (18, 12), (19, 12), (20, 12), (21, 12),
            (5, 4), (6, 4), (7, 4),
            (11, 5), (12, 5), (13, 5), (14, 5),
            (18, 5), (19, 5), (20, 5), (21, 5),
            (25, 4), (26, 4), (27, 4),
        ],
        "max_amount": 2,
        "disappearance_probability": 0.0,
    },
    protocol.ObjectType.Greenshard: {
        "probability": 0.005,
        "area": [
            (11, 12), (12, 12), (13, 12), (14, 12), (15, 12), (16, 12), (17, 12), (18, 12), (19, 12), (20, 12), (21, 12),
        ],
        "max_amount": 1,
        "disappearance_probability": 0.0,
    }
}



class Object:

    id: int
    map_id: int
    pos_x: int
    pos_y: int
    object_type: protocol.ObjectType


    def __init__(self, id, map_id, pos_x, pos_y, object_type):

        self.id = id
        self.map_id = map_id
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.object_type = object_type


    @staticmethod
    def generate_id():
        while (x := random.getrandbits(16)) == 0:
            continue
        return x



class WorldObjects:

    objects: dict[int, Object]
    lock: threading.RLock


    def __init__(self):

        self.objects = {}
        self.lock = threading.RLock()


    def print_objects(self) -> None:
        # For debugging purposes
        with self.lock:
            print(f"Number of objects: {len(self.objects)}")
            for obj in self.objects.values():
                print(f"id=0x{obj.id:04x}, pos_x={obj.pos_x}, pos_y={obj.pos_y}, type={obj.object_type}")
            print()


    def count_objects_by_type(self, object_type: protocol.ObjectType) -> int:

        with self.lock:

            return sum(1 for obj in self.objects.values() if obj.object_type == object_type)


    def is_position_occupied_by_object(self, pos_x, pos_y) -> bool:

        with self.lock:

            for obj in self.objects.values():
                if obj.pos_x == pos_x and obj.pos_y == pos_y:
                    return True

            return False


    def get_object_by_id(self, object_id: int) -> Object | None:

        with self.lock:

            if object_id not in self.objects:
                return None

            return self.objects[object_id]


    def remove_object(self, object_id: int) -> None:

        with self.lock:

            if object_id in self.objects:
                del self.objects[object_id]


    def generate_available_object_id(self) -> int | None:

        with self.lock:

            if len(self.objects) >= MAX_OBJECTS: return None

            while (object_id := Object.generate_id()) in self.objects:
                continue

            return object_id


    def add_object_at_random_position(self, map_id: int, object_type: protocol.ObjectType) -> None:

        with self.lock:

            object_id = self.generate_available_object_id()
            if object_id is None: return

            while True:
                pos_x, pos_y = random.choice(OBJECTS[object_type]["area"])
                if not self.is_position_occupied_by_object(pos_x, pos_y):
                    break

            self.objects[object_id] = Object(object_id, map_id, pos_x, pos_y, object_type)


    def add_random_objects(self, map_id: int) -> None:

        with self.lock:

            for object_type in OBJECTS:

                if random.random() >= OBJECTS[object_type]["probability"]: continue
                if self.count_objects_by_type(object_type) >= OBJECTS[object_type]["max_amount"]: continue

                self.add_object_at_random_position(map_id, object_type)


    def delete_random_object(self) -> None:

        with self.lock:

            if len(self.objects) == 0: return

            object_id = random.choice(list(self.objects.keys()))
            object_type = self.objects[object_id].object_type

            if random.random() >= OBJECTS[object_type]["disappearance_probability"]: return

            self.remove_object(object_id)



world_objects: WorldObjects = WorldObjects()
world_objects.add_object_at_random_position(MAIN_ARENA_ID, protocol.ObjectType.Greenshard)
