from construct import ConstructError, Container, EnumIntegerString
from typing import Tuple, Callable, TypeAlias

import socket
import sys
import time
import threading

from player import Player, player_db, player_lock
from map import world_objects, MAIN_ARENA_ID
import protocol



LICENSE_KEY = b"PR2YU5CZGCYMS272GLZ1WA43W7P44I7S"

MAX_PLAYERS = 8
MAX_TIMEOUT = 15.0
UPDATE_PERIOD = 1/45.0  # Try to be slightly faster than TARGET_FPS in clients


sock: socket.socket
Addr: TypeAlias = Tuple[str, int]
PduHandler: TypeAlias = Callable[[Addr, Container, Player | None], None]



# Sending functions


def send_pdu(addr: Addr, cmd: protocol.Cmd, body: dict) -> None:

    global sock

    serialized_body = protocol.commands[cmd].build(body)
    # print(f"Sending {cmd!r}, body ({len(serialized_body)} bytes): {serialized_body.hex()}")

    assert 0 <= len(serialized_body) <= 0xFFFF

    serialized_pdu = protocol.PDU.build({
        "cmdId": cmd,
        "bodySize": len(serialized_body),
        "token": b"\x00" * 16,
        "body": serialized_body,
    })

    sock.sendto(serialized_pdu, addr)


def broadcast_pdu(cmd: protocol.Cmd, body: dict) -> None:

    with player_lock:
        addr_list = [player.addr for player in player_db.values()]

    for addr in addr_list:
        send_pdu(addr, cmd, body)


def send_error(addr: Addr, error: str) -> None:
    send_pdu(addr, protocol.Cmd.ERROR_PDU, {"error": error})



# Receiving functions


def handle_authentication_pdu(addr: Addr, body: Container, _player: Player | None) -> None:

    if body.license_key != LICENSE_KEY:
        send_error(addr, "Invalid license key")
        return

    if (username_size := body.username.find(b"\x00")) <= 0:
        send_error(addr, "Invalid username")
        return

    try:
        username = body.username[:username_size].decode()
    except UnicodeDecodeError:
        send_error(addr, "Invalid username")
        return

    with player_lock:

        if len(player_db) >= MAX_PLAYERS:
            send_error(addr, "Server reached player limit")
            return

        while (player_id := Player.generate_id()) in player_db:
            continue

        player_token = Player.generate_token()

        blue_count = len(Player.get_players_by_team(protocol.Team.Blue))
        red_count = len(Player.get_players_by_team(protocol.Team.Red))

        if blue_count > red_count: team_color = protocol.Team.Red
        else: team_color = protocol.Team.Blue

        player_db[player_id] = Player(player_id, addr, team_color, username, player_token)
        total_players = blue_count + red_count + 1

    print(f"[+] New player '{username}' joined (0x{player_id:04x}) ({'red' if team_color == protocol.Team.Red else 'blue'} team) (total: {total_players}p)")
    broadcast_pdu(protocol.Cmd.TEXT_MESSAGE_PDU, {"text": f"{username} joined"})

    send_pdu(addr, protocol.Cmd.AUTHENTICATION_RESPONSE_PDU, {
        "player_id": player_id,
        "token": player_token,
        "team": team_color,
    })


def handle_heartbeat_pdu(_addr: Addr, _body: Container, player: Player | None) -> None:
    with player_lock:
        player.last_heartbeat = time.time()


def handle_client_player_info_pdu(addr: Addr, body: Container, player: Player | None) -> None:

    if not isinstance(body.facing, EnumIntegerString):
        send_error(addr, "Invalid facing value")
        return

    with player_lock:

        # No specific validation of user-provided coordinates, you're free to cheat... =)
        player.map_id = body.map_id
        player.x = body.x
        player.y = body.y

        player.facing = body.facing
        player.is_afk = bool(body.is_afk)
        player.shielded = bool(body.shielded)
        player.greenshard_hp = body.greenshard_hp

        if (username_size := body.name.find(b"\x00")) <= 0:
            send_error(addr, "Invalid username")
            return

        try:
            player.name = body.name[:username_size].decode()
        except UnicodeDecodeError:
            send_error(addr, "Invalid username")
            return

        if (last_attacker_name_size := body.last_attacker.find(b"\x00")) < 0:
            send_error(addr, "Invalid last attacker name")
            return

        try:
            player.last_attacker = body.last_attacker[:last_attacker_name_size].decode()
        except UnicodeDecodeError:
            send_error(addr, "Invalid last attacker name")
            return


def handle_object_seizing_pdu(addr: Addr, body: Container, player: Player | None) -> None:

    with world_objects.lock:

        if (object_ := world_objects.get_object_by_id(body.object_id)) is None:
            send_error(addr, "Invalid object id, maybe object just disappeared or was taken")
            return

        world_objects.remove_object(object_.id)

    match object_.object_type:

        case protocol.ObjectType.Heart:
            with player_lock:
                player.heal(5)

        case protocol.ObjectType.Greenshard:
            # Shield effect is managed client-side, we don't have much to do
            pass


def handle_client_attack_pdu(addr: Addr, body: Container, player: Player | None) -> None:

    death = False
    respawn_map_id, respawn_x, respawn_y = 0, 0, 0

    with player_lock:

        if body.victim_player_id not in player_db:
            send_error(addr, "Invalid target player id")
            return

        victim = player_db[body.victim_player_id]

        if not victim.shielded:
            victim.receive_damage(body.attack_damage)

        victim_name, victim_id = victim.name, victim.id
        perpetrator_name, perpetrator_id = player.name, player.id

        if victim.hp == 0:
            death = True
            respawn_map_id, respawn_x, respawn_y = victim.revive()

    broadcast_pdu(protocol.Cmd.SERVER_ATTACK_PDU, {
        "perpetrator_id": perpetrator_id,
        "victim_id": victim_id,
        "damage": body.attack_damage,
        "is_dead": int(death),
        "respawn_map_id": respawn_map_id,
        "respawn_x": respawn_x,
        "respawn_y": respawn_y,
    })

    if death:
        broadcast_pdu(protocol.Cmd.TEXT_MESSAGE_PDU, {"text": f"{perpetrator_name} killed {victim_name}"})


def handle_client_disconnect_pdu(_addr: Addr, _body: Container, player: Player | None) -> None:

    with player_lock:
        player_id = player.id
        player_name = player.name
        print(f"[+] Player '{player_name}' disconnected (0x{player_id:04x})")
        del player_db[player_id]
        del player

    broadcast_pdu(protocol.Cmd.TEXT_MESSAGE_PDU, {"text": f"{player_name} left"})
    broadcast_pdu(protocol.Cmd.SERVER_DISCONNECT_PDU, {"player_id": player_id})


def handle_chat_pdu(_addr: Addr, body: Container, player: Player | None) -> None:

    with player_lock:
        player_name = player.name

    print(f"[+] Player '{player_name}' said: {body.text!r}")
    broadcast_pdu(protocol.Cmd.CHAT_PDU, {"text": body.text})



# Updater thread


def check_timeouts() -> None:

    now = time.time()
    timeouts = []

    with player_lock:
        for player in list(player_db.values()):
            if now - player.last_heartbeat > MAX_TIMEOUT:
                # Player timed out: kick them
                timeouts.append((player.id, player.name))
                del player_db[player.id]
                del player

    for player_id, player_name in timeouts:
        print(f"[+] Player '{player_name}' timed out (0x{player_id:04x})")
        broadcast_pdu(protocol.Cmd.TEXT_MESSAGE_PDU, {"text": f"{player_name} timed out"})
        broadcast_pdu(protocol.Cmd.SERVER_DISCONNECT_PDU, {"player_id": player_id})


def send_server_players_info() -> None:

    infos = []

    with player_lock:
        for player in list(player_db.values()):
            infos.append({
                "player_id": player.id,
                "team": player.team,
                "name": player.name.encode().ljust(16, b"\x00"),
                "facing": player.facing,
                "hp": player.hp,
                "map_id": player.map_id,
                "x": player.x,
                "y": player.y,
                "is_afk": int(player.is_afk),
                "shielded": int(player.shielded),
                "greenshard_hp": player.greenshard_hp,
                "last_attacker": player.last_attacker.encode().ljust(16, b"\x00"),
            })

    for info in infos:
        broadcast_pdu(protocol.Cmd.SERVER_PLAYER_INFO_PDU, info)


def send_map_state() -> None:

    objects: list = []

    with world_objects.lock:
        for object_ in world_objects.objects.values():
            objects.append({
                "object_id": object_.id,
                "map_id": object_.map_id,
                "pos_x": object_.pos_x,
                "pos_y": object_.pos_y,
                "object_type": object_.object_type,
            })

    broadcast_pdu(protocol.Cmd.MAP_STATE_PDU, {
        "n_objects": len(objects),
        "objects": objects,
    })


def updater() -> None:

    stop = threading.Event()

    while not stop.is_set():

        stop.wait(UPDATE_PERIOD)

        world_objects.add_random_objects(MAIN_ARENA_ID)
        # world_objects.delete_random_object()

        check_timeouts()
        send_server_players_info()
        send_map_state()

        # world_objects.print_objects()


# Main

def main() -> None:

    global sock

    data: bytes
    addr: Addr

    if len(sys.argv) != 3:
        print("[x] Usage: python server.py <bind_ip> <bind_port>")
        exit(1)

    ip: str = sys.argv[1]
    port: int = int(sys.argv[2])

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((ip, port))

    print(f"[+] Listening on {ip}:{port} (UDP)")

    timer_updater = threading.Timer(0, updater)
    timer_updater.daemon = True
    timer_updater.start()

    while True:

        data, addr = sock.recvfrom(65536)

        # print(f"[+] Received {len(data)} bytes from {addr}: {data.hex()}")

        if len(data) < 0x14:
            send_error(addr, "Invalid header size, packet ignored")
            continue

        try:
            pdu: Container = protocol.PDU.parse(data)
        except ConstructError as e:
            send_error(addr, "Malformed PDU, packet ignored")
            print(e)
            print(data.hex())
            continue

        if pdu.cmdId not in pdu_handlers:
            send_error(addr, "Unknown or unsupported command, packet ignored")
            continue

        try:
            body: Container = protocol.commands[pdu.cmdId].parse(bytes(pdu.body))
        except ConstructError:
            send_error(addr, "Malformed body, packet ignored")
            continue

        player: Player | None = None

        if pdu.cmdId != protocol.Cmd.AUTHENTICATION_PDU:
            with player_lock:
                if (player := Player.get_player_by_token(pdu.token)) is None:
                    send_error(addr, "Invalid token, packet ignored")
                    continue

        pdu_handlers[pdu.cmdId](addr, body, player)

    timer_updater.join()
    sock.close()


pdu_handlers: dict[int, PduHandler] = {
    protocol.Cmd.AUTHENTICATION_PDU: handle_authentication_pdu,
    protocol.Cmd.HEARTBEAT_PDU: handle_heartbeat_pdu,
    protocol.Cmd.CLIENT_PLAYER_INFO_PDU: handle_client_player_info_pdu,
    protocol.Cmd.CLIENT_ATTACK_PDU: handle_client_attack_pdu,
    protocol.Cmd.OBJECT_SEIZING_PDU: handle_object_seizing_pdu,
    protocol.Cmd.CLIENT_DISCONNECT_PDU: handle_client_disconnect_pdu,
    protocol.Cmd.CHAT_PDU: handle_chat_pdu,
}


if __name__ == "__main__":
    main()
