#include "message.h"

#include "openssl/evp.h"

#include "csr.h"
#include "error.h"
#include <stddef.h>

size_t format(char **message, char *username, size_t username_sz,
              char *recipient, size_t recipient_sz, char *content,
              size_t content_sz) {
  if (*message != NULL) {
    hsm_error(ERROR_INTERNAL);
  }

  // message_sz = username_sz + recipient_sz + content_sz
  // + len("from:\to:\content:")
  size_t message_sz = username_sz + recipient_sz + content_sz + 18;
  *message = malloc(message_sz + 1); // + 1 for \0
  snprintf(*message, message_sz + 1, "from:%s\\to:%s\\content:%s", username,
           recipient, content);

  return message_sz;
}

size_t encode(char **out, unsigned char *in, size_t in_sz) {
  if (*out != NULL) {
    hsm_error(ERROR_INTERNAL);
  }

  size_t out_sz = (in_sz + 2) / 3 * 4;
  *out = malloc(out_sz + 1); // + 1 for \0

  int encoded_sz = EVP_EncodeBlock((void *)*out, in, in_sz);
  if (encoded_sz < 0 || out_sz != (size_t)encoded_sz) {
    hsm_error(ERROR_OPENSSL);
  }

  return out_sz;
}

void sign(char *out, unsigned char *message, size_t message_sz) {
  unsigned char key[SIG_SZ];
  for (int i = 0; i < SIG_SZ / 4; ++i) {
    unsigned long key32 = csrr[i]();
    key[4 * i + 0] = key32 & 0xFF;
    key[4 * i + 1] = key32 >> 8 & 0xFF;
    key[4 * i + 2] = key32 >> 16 & 0xFF;
    key[4 * i + 3] = key32 >> 24 & 0xFF;
  }

  EVP_MAC *mac = EVP_MAC_fetch(NULL, "HMAC", NULL);
  if (mac == NULL) {
    hsm_error(ERROR_OPENSSL);
  }

  EVP_MAC_CTX *mac_ctx = EVP_MAC_CTX_new(mac);
  if (mac_ctx == NULL) {
    hsm_error(ERROR_OPENSSL);
  }
  EVP_MAC_free(mac);

  OSSL_PARAM params[2];
  params[0] = OSSL_PARAM_construct_utf8_string("digest", "SHA256", 0);
  params[1] = OSSL_PARAM_construct_end();
  if (EVP_MAC_init(mac_ctx, key, SIG_SZ, params) != 1) {
    hsm_error(ERROR_OPENSSL);
  }

  if (EVP_MAC_update(mac_ctx, message, message_sz) != 1) {
    hsm_error(ERROR_OPENSSL);
  }

  unsigned char signature[SIG_SZ];
  if (EVP_MAC_final(mac_ctx, signature, NULL, SIG_SZ) != 1) {
    hsm_error(ERROR_OPENSSL);
  }
  EVP_MAC_CTX_free(mac_ctx);

  if (EVP_EncodeBlock((void *)out, signature, SIG_SZ) != SIG_SZ_B64) {
    hsm_error(ERROR_OPENSSL);
  }
}

int verify(char *claim_b64, unsigned char *message, size_t message_sz) {
  char signature_b64[SIG_SZ_B64 + 1]; // + 1 for /0
  sign(signature_b64, message, message_sz);

  if (CRYPTO_memcmp(signature_b64, claim_b64, SIG_SZ_B64) != 0) {
    return 0; // Failure
  }
  return 1; // Success
}
