import base64
import json

import filelock
import flask

import hsmm

app = flask.Flask(__name__)

LEN_DATA_MAX = 512
LEN_SIG = 44
MSGS = "msgs.json"
MSGS_LOCK = "msgs.json.lock"
VERSION = "version"
VERSION_LOCK = "version.lock"

KNOWN_USERS = [
    "@mafiaBOSS",
    "@mafiaBRO",
    "@mafiaDEV",
    "@mafiaTECHNICIAN",
    "@mafiaTEST",
]

USER_NAMES = {
    "@mafiaBOSS": "EMERALD",
    "@mafiaBRO": "BRO",
    "@mafiaDEV": "DEV",
    "@mafiaTECHNICIAN": "TECHNOMAFIOSO",
    "@mafiaTEST": "TEST",
}

USER_DESCRIPTIONS = {
    "@mafiaBOSS": "CMO (Chief Mafia Officer)",
    "@mafiaBRO": "Apostle of the Green Shard Order",
    "@mafiaDEV": "The Ultimate Developer",
    "@mafiaTECHNICIAN": "Tesla goes vroom vroom",
    "@mafiaTEST": "Test user, password is the default one (OnlyRequiredToSignAndSendMessages)",
}


@app.get("/users")
def users_get():
    users = []
    for id in KNOWN_USERS:
        current_user = {}
        current_user["id"] = id
        name = USER_NAMES.get(id, None)
        current_user["name"] = name if name is not None else ""
        description = USER_DESCRIPTIONS.get(id, None)
        current_user["description"] = description if description is not None else ""
        users.append(current_user)
    return {"result": {"users": users}}


def read_version():
    try:
        with filelock.FileLock(VERSION_LOCK):
            with open(VERSION) as f:
                return int(f.readline())
    except FileNotFoundError:
        return 0


def read_msgs():
    try:
        with filelock.FileLock(VERSION_LOCK):
            with open(VERSION) as f:
                version = int(f.readline())
            with filelock.FileLock(MSGS_LOCK):
                with open(MSGS) as f:
                    msgs = json.load(f)
            return (version, msgs)
    except FileNotFoundError:
        return (0, {"msgs": []})


def write_msgs(version, msgs):
    with filelock.FileLock(VERSION_LOCK):
        try:
            with open(VERSION, "r") as f:
                file_version = int(f.readline())
            if version != file_version:
                flask.abort(503, "Mismatched messages versions")
        except FileNotFoundError:
            pass
        version += 1
        with open(VERSION, "w") as f:
            f.write(str(version))
        with filelock.FileLock(MSGS_LOCK):
            with open(MSGS, "w") as f:
                json.dump(msgs, f)
                return version


@app.get("/version")
def version_get():
    version = read_version()
    return {"version": version}


@app.get("/msgs")
def msgs_get():
    version, msgs = read_msgs()
    return {"version": version, "result": msgs}


@app.post("/msgs")
def msgs_post():
    data_b64 = flask.request.form["data"]
    sig_b64 = flask.request.form["sig"]
    if LEN_DATA_MAX < len(data_b64) or LEN_SIG != len(sig_b64):
        flask.abort(400, "Bad lengths")
    try:
        data = base64.b64decode(data_b64)
    except Exception:
        flask.abort(400, "Bad base64 in data parameter")
    try:
        verify = hsmm.verify(data, sig_b64)
    except hsmm.HSMMException:
        flask.abort(400, "HSMM rejected the message")
    if not verify:
        flask.abort(403, "HSMM rejected the signature")
    msg = {
        "from": None,
        "to": None,
        "content": None,
    }
    msg_key = None
    buf_str = ""
    for i in data.decode("utf-8", "ignore"):
        if i == ":":
            msg_key = buf_str
            buf_str = ""
        elif i == "\\":
            if msg_key in msg:
                msg[msg_key] = buf_str
            msg_key = None
            buf_str = ""
        else:
            buf_str += i
    if msg_key in msg:
        msg[msg_key] = buf_str
    for i in msg.values():
        if i is None:
            flask.abort(400, f"Missing field in message: {i}")
    if (msg["from"] not in KNOWN_USERS) or (msg["to"] not in KNOWN_USERS):
        flask.abort(403, "Unknown user")
    if (msg["from"] == "@mafiaTEST"):
        if (msg["to"] != "@mafiaDEV"):
            flask.abort(403, "Test user is restricted, it can only send messages to @mafiaDEV")
        msg["content"] = "Test user is restricted, the message was correctly sent but the content is redacted"
    else:
        try:
            msg["content"] = base64.b64decode(msg["content"]).decode("utf-8", "ignore")
        except Exception:
            flask.abort(400, "Bad base64 in message content")
    version, msgs = read_msgs()
    msgs["msgs"].append(msg)
    version = write_msgs(version, msgs)
    return {"version": version, "result": msgs}
