from core.core30_context.context_dependency_graph import context_producer, context_dependencies
from core.core30_context.context import Context
from action.utils.local_state import Store

from datetime import datetime, timezone
from pydantic import BaseModel
from typing import List, Dict, Literal


class UnitCryptoMessageModel(BaseModel):
    type: Literal['ENC'] = 'ENC'
    length: int
    enc_bytes: str
    to_server: bool
    execution_context: str
    emitted_at: datetime


class UnitErrorMessageModel(BaseModel):
    type: Literal['ERROR'] = 'ERROR'
    length: int
    enc_bytes: str
    to_server: bool
    execution_context: str
    emitted_at: datetime


UnitMessageModel = UnitCryptoMessageModel | UnitErrorMessageModel


def dict_to_model(d):
    try:
        return UnitCryptoMessageModel(**d)
    except:
        return UnitErrorMessageModel(**d)


class DecryptedBlockModel(BaseModel):
    base_id: int
    block_id: int
    decrypted: str


class DecryptedMessageModel(BaseModel):
    base_id: int
    decrypted: str


class ParsedDecryptedUnitMessage(BaseModel):
    base_id: int
    message_type: int


class UnitMessageWithAnswerModel(BaseModel):
    message: UnitMessageModel
    message_id: int
    answer: UnitMessageModel | None = None


@context_producer(('.exploit.step1.traces', List[UnitMessageModel]),
                  ('.exploit.step1.decrypted_blocks', List[DecryptedBlockModel]),
                  ('.exploit.step1.decrypted_messages', Dict[int, DecryptedMessageModel]))
@context_dependencies(('.state.store', Store))
def init_store(ctxt: Context):
    current_store = ctxt['state']['store']
    if 'collected_traces' not in current_store:
        current_store['collected_traces'] = []
    if 'decrypted_messages' not in current_store:
        current_store['decrypted_messages'] = {}
    if 'decrypted_blocks' not in current_store:
        current_store['decrypted_blocks'] = []
    ctxt.setdefault('exploit', {}).setdefault('step1', {})['traces'] = current_store['collected_traces']
    ctxt['exploit']['step1']['decrypted_blocks'] = current_store['decrypted_blocks']
    ctxt['exploit']['step1']['decrypted_messages'] = current_store['decrypted_messages']


@context_dependencies(('.exploit.step1.traces', List[UnitMessageModel]), ('.state.store', Store))
def receive_message(ctxt: Context, data, from_client, execution_context):
    traces = ctxt['exploit']['step1']['traces']
    if data[0] == b'\x00':
        new_message = UnitErrorMessageModel(length=len(data), enc_bytes=data[1:].decode(), to_server=from_client,
                                            execution_context=execution_context, emitted_at=datetime.now(timezone.utc))
    else:
        new_message = UnitCryptoMessageModel(length=len(data), enc_bytes=data.hex(), to_server=from_client,
                                             execution_context=execution_context, emitted_at=datetime.now(timezone.utc))
    traces.append(new_message.dict())
    ctxt['state']['store']['collected_traces'] = traces  # trigger save
    return len(traces)-1, new_message


@context_dependencies(('.exploit.step1.traces', List[UnitMessageModel]))
def get_all_messages(ctxt: Context):
    return [dict_to_model(m) for m in ctxt['exploit']['step1']['traces']]


def message_with_answer_if_any(all_messages, m, i):
    if i + 1 < len(all_messages) and all_messages[i + 1].execution_context == all_messages[i].execution_context and \
            all_messages[i + 1].to_server is not all_messages[i].to_server:
        return UnitMessageWithAnswerModel(message=m, answer=all_messages[i + 1], message_id=i)
    else:
        return UnitMessageWithAnswerModel(message=m, message_id=i)

def get_all_messages_with_answer():
    all_messages = get_all_messages()
    return [
        message_with_answer_if_any(all_messages, m, i)
        for i, m in enumerate(all_messages)
    ]


@context_dependencies(('.exploit.step1.traces', List[UnitMessageModel]),
                      ('.exploit.step1.decrypted_messages', Dict[int, DecryptedMessageModel]))
def get_all_unknown_messages(ctxt: Context):
    return [dict_to_model(m) for i, m in enumerate(ctxt['exploit']['step1']['traces'])
            if i not in ctxt['exploit']['step1']['decrypted_messages']]

@context_dependencies(('.exploit.step1.decrypted_messages', Dict[int, DecryptedMessageModel]))
def get_all_unknown_messages_with_answers(ctxt: Context):
    all_messages = get_all_messages()
    return [
        message_with_answer_if_any(all_messages, m, i)
        for i, m in enumerate(all_messages) if i not in ctxt['exploit']['step1']['decrypted_messages']
                                               and m.type == 'ENC'
    ]

@context_dependencies(('.exploit.step1.decrypted_messages', Dict[int, DecryptedMessageModel]))
def get_all_known_messages_with_answers(ctxt: Context):
    all_messages = get_all_messages()
    return [
        message_with_answer_if_any(all_messages, m, i)
        for i, m in enumerate(all_messages) if i in ctxt['exploit']['step1']['decrypted_messages'] and m.type == 'ENC'
    ]


@context_dependencies(('.exploit.step1.decrypted_blocks', List[DecryptedBlockModel]))
def get_all_decrypted_blocks(ctxt: Context):
    return [DecryptedBlockModel(**b) for b in ctxt['exploit']['step1']['decrypted_blocks']]


@context_dependencies(('.exploit.step1.decrypted_messages', Dict[int, DecryptedMessageModel]))
def get_all_decrypted_messages(ctxt: Context):
    return ctxt['exploit']['step1']['decrypted_messages']


@context_producer(('.exploit.step1.decrypted_matches', Dict[str, str]))
def get_all_decrypted_matches(ctxt: Context) -> Dict[str, str]:
    traces = get_all_messages()
    dec_blocks = get_all_decrypted_blocks()
    corresp = {
        traces[b.base_id].enc_bytes[2 + 0x20 * (b.block_id + 1): 0x22 + 0x20 * (b.block_id + 1)]: b.decrypted
        for b in dec_blocks if b.base_id < len(traces)
    }
    ctxt['exploit']['step1']['decrypted_matches'] = corresp
    print("GET ALL DEC MATCHES")
    print(len(corresp))
    return corresp


@context_dependencies(('.exploit.step1.decrypted_blocks', List[DecryptedBlockModel]), ('.state.store', Store))
def decryption_success(ctxt: Context, trace_id: int, block_in: int, block_out: bytes):
    blocks = ctxt['exploit']['step1']['decrypted_blocks']
    blocks.append(DecryptedBlockModel(base_id=trace_id, block_id=block_in, decrypted=block_out.hex()).dict())
    ctxt['state']['store']['decrypted_blocks'] = blocks  # trigger save


@context_dependencies(('.exploit.step1.decrypted_blocks', List[DecryptedBlockModel]), ('.state.store', Store))
def refresh_decrypted(ctxt: Context):
    all_messages = get_all_messages()
    all_matches = get_all_decrypted_matches()
    new_decrypted = {}

    for idx, m in enumerate(all_messages):
        msg = b''
        if m.type == 'ENC':
            decoded = bytes.fromhex(m.enc_bytes)
            for i in range(1, len(decoded) // 0x10):
                chunk = decoded[1 + i * 0x10: 1 + (i + 1) * 0x10].hex()
                if chunk not in all_matches:
                    break
                else:
                    msg += bytes.fromhex(all_matches[chunk])
        if len(msg) + 0x11 == len(decoded):
            print(msg)
            new_decrypted[idx] = DecryptedMessageModel(base_id=idx, decrypted=msg.hex())

    ctxt['exploit']['step1']['decrypted_messages'] = new_decrypted
    ctxt['state']['store']['decrypted_messages'] = {k: new_decrypted[k].dict() for k in new_decrypted}  # trigger save
    return new_decrypted
